@@ -1,16 +1,26 @@
 //! Module processing functionality.
 
+mod index;
 mod interpolator;
 mod layouter;
 mod namer;
 mod terminator;
 mod typifier;
 
+pub use index::IndexableLength;
 pub use layouter::{Alignment, InvalidBaseType, Layouter, TypeLayout};
 pub use namer::{EntryPointIndex, NameKey, Namer};
 pub use terminator::ensure_block_returns;
 pub use typifier::{ResolveContext, ResolveError, TypeResolution};
 
+#[derive(Clone, Debug, thiserror::Error, PartialEq)]
+pub enum ProcError {
+    #[error("type is not indexable, and has no length (validation error)")]
+    TypeNotIndexable,
+    #[error("array length is wrong kind of constant (validation error)")]
+    InvalidArraySizeConstant(crate::Handle<crate::Constant>),
+}
+
 impl From<super::StorageFormat> for super::ScalarKind {
     fn from(format: super::StorageFormat) -> Self {
         use super::{ScalarKind as Sk, StorageFormat as Sf};
@@ -75,19 +85,31 @@ impl super::TypeInner {
         }
     }
 
+    pub fn pointer_class(&self) -> Option<crate::StorageClass> {
+        match *self {
+            Self::Pointer { class, .. } => Some(class),
+            Self::ValuePointer { class, .. } => Some(class),
+            _ => None,
+        }
+    }
+
     pub fn span(&self, constants: &super::Arena<super::Constant>) -> u32 {
         match *self {
-            Self::Scalar { kind: _, width } => width as u32,
+            Self::Scalar { kind: _, width } | Self::Atomic { kind: _, width } => width as u32,
             Self::Vector {
                 size,
                 kind: _,
                 width,
             } => (size as u8 * width) as u32,
+            // matrices are treated as arrays of aligned columns
             Self::Matrix {
                 columns,
                 rows,
                 width,
-            } => (columns as u8 * rows as u8 * width) as u32,
+            } => {
+                let aligned_rows = if rows > crate::VectorSize::Bi { 4 } else { 2 };
+                columns as u32 * aligned_rows * width as u32
+            }
             Self::Pointer { .. } | Self::ValuePointer { .. } => POINTER_SPAN,
             Self::Array {
                 base: _,
@@ -217,7 +239,19 @@ impl crate::SampleLevel {
 }
 
 impl crate::Constant {
-    pub fn to_array_length(&self) -> Option<u32> {
+    /// Interpret this constant as an array length, and return it as a `u32`.
+    ///
+    /// Ignore any specialization available for this constant; return its
+    /// unspecialized value.
+    ///
+    /// If the constant has an inappropriate kind (non-scalar or non-integer) or
+    /// value (negative, out of range for u32), return `None`. This usually
+    /// indicates an error, but only the caller has enough information to report
+    /// the error helpfully: in back ends, it's a validation error, but in front
+    /// ends, it may indicate ill-formed input (for example, a SPIR-V
+    /// `OpArrayType` referring to an inappropriate `OpConstant`). So we return
+    /// `Option` and let the caller sort things out.
+    pub(crate) fn to_array_length(&self) -> Option<u32> {
         use std::convert::TryInto;
         match self.inner {
             crate::ConstantInner::Scalar { value, width: _ } => match value {
@@ -288,3 +322,17 @@ impl super::SwizzleComponent {
         }
     }
 }
+
+#[test]
+fn test_matrix_size() {
+    let constants = crate::Arena::new();
+    assert_eq!(
+        crate::TypeInner::Matrix {
+            columns: crate::VectorSize::Tri,
+            rows: crate::VectorSize::Tri,
+            width: 4
+        }
+        .span(&constants),
+        48
+    );
+}
