@@ -39,6 +39,8 @@ pub fn ensure_block_returns(block: &mut crate::Block) {
         | Some(&mut S::Store { .. })
         | Some(&mut S::ImageStore { .. })
         | Some(&mut S::Call { .. })
-        | None => block.push(S::Return { value: None }),
+        | Some(&mut S::Atomic { .. })
+        | Some(&mut S::Barrier(_))
+        | None => block.push(S::Return { value: None }, Default::default()),
     }
 }
